import 'package:contentsquare/src/constants/gesture_targetable_widgets.dart';
import 'package:contentsquare/src/modules/screengraph/screengraph_node.dart';
import 'package:contentsquare/src/modules/screengraph/screengraph_utils.dart';

import 'package:flutter/widgets.dart';

class _ScreengraphNodeOrChildren {
  final Map? node;
  final List? children;

  _ScreengraphNodeOrChildren({this.node, this.children})
      : assert(node != null || children != null);
}

ScreengraphNode _getRootNode(String rootPath) {
  final rootElement = WidgetsBinding.instance!.renderViewElement!;
  return ScreengraphNode(
    element: rootElement,
    rootPath: rootPath,
  );
}

/// Keep track of all element paths
mixin _ElementPathsTracker {
  final Map<String, int> _elementPathOccurrences = {};

  void _updateElementPathOccurrences(String newPath) {
    final occurence = _elementPathOccurrences[newPath];
    if (occurence != null) {
      _elementPathOccurrences[newPath] = occurence + 1;
    } else {
      _elementPathOccurrences[newPath] = 0;
    }
  }

  int getNumberOfOccurencesForElementPath(String elementPath) {
    _updateElementPathOccurrences(elementPath);
    final occurence = _elementPathOccurrences[elementPath];

    /// _elementPathOccurrences[path] is updated by _updateElementPathOccurrences and should not end up null by the time it's called
    return occurence ?? 0;
  }
}

/// Scan screengraph for [Element] aroung a given position
class ScreengraphScanner with _ElementPathsTracker {
  /// Scan the element tree for candidate elements around the given position
  String scanRootForCandidateElementsAtPosition(Offset position) {
    final List<String> targetedElementsPaths = [];
    _searchForNodeAtPosition(
      position,
      stopWhere: (node) {
        /// If the element is one of the "targetable elements", we don't need to go further down the tree
        return _isWidgetTargetable(node.element);
      },
      onCandidateNodeVisited: (ScreengraphNode node) {
        targetedElementsPaths.add(node.path);
      },
    );
    return targetedElementsPaths.last;
  }

  void _searchForNodeAtPosition(Offset position,
      {required Function(ScreengraphNode node) onCandidateNodeVisited,
      Function(ScreengraphNode node)? stopWhere}) {
    _exploreSubtreeAroundPosition(
      position,
      _getRootNode(""),
      onCandidateNodeVisited: onCandidateNodeVisited,
      stopWhere: stopWhere,
    );
  }

  /// Explore current screengraph given a [position] and call the provided [onCandidateNodeVisited] on each candidate visited
  /// It ignores sections of the tree with frame (x,y, width, heigth) that does not match the given [position]
  /// Candidates are widgets with a frame matching the given [position] and that are either "leaf" of the element tree (without children)
  /// or matches the optionally provided [stopWhere] condition
  /// Once [stopWhere] is reached, the search stops and immediatyl calls the [onCandidateNodeVisited] callback.
  /// It does not go any further down the element tree
  void _exploreSubtreeAroundPosition(Offset position, ScreengraphNode node,
      {required Function(ScreengraphNode node) onCandidateNodeVisited,
      Function(ScreengraphNode node)? stopWhere}) {
    node.appendPositionalIndexToPath(
        getNumberOfOccurencesForElementPath(node.path));

    if (!node.frame.contains(position)) {
      return;
    }

    /// If the [stopWhere] condition has been reached, call [onLastNodeVisited] callback and exit
    if (stopWhere != null && stopWhere(node)) {
      onCandidateNodeVisited(node);
      return;
    }

    /// Otherwise, checks if the element is a leaf
    final element = node.element;
    bool isLeafElement = element.toDiagnosticsNode().getChildren().isEmpty;
    if (isLeafElement) {
      onCandidateNodeVisited(node);
    }

    /// Else, keep on visiting the tree deeper
    node.visitChildren(
      (ScreengraphNode node) => _exploreSubtreeAroundPosition(position, node,
          stopWhere: stopWhere, onCandidateNodeVisited: onCandidateNodeVisited),
    );
  }

  static bool _isWidgetTargetable(Element element) {
    return gestureTargetableWidgets
        .contains(element.widget.runtimeType.toString());
  }
}

/// Generate a screengraph of the current Element Tree
class ScreengraphGenerator with _ElementPathsTracker {
  /// Generate a screengraph of the current Element tree and prefix it with the given [rootPath]
  Future<List> generateScreengraph(String rootPath,
      {required double pixelRatio}) async {
    return [
      (await _generateSubScreengraph(_getRootNode(rootPath),
              pixelRatio: pixelRatio))
          .node
    ];
  }

  Future<_ScreengraphNodeOrChildren> _generateSubScreengraph(
      ScreengraphNode node,
      {required double pixelRatio}) async {
    node.appendPositionalIndexToPath(
        getNumberOfOccurencesForElementPath(node.path));

    List<Map> children = [];
    for (DiagnosticsNode childElement
        in node.element.toDiagnosticsNode().getChildren()) {
      _ScreengraphNodeOrChildren child = await _generateSubScreengraph(
          ScreengraphNode(element: childElement.value as Element, parent: node),
          pixelRatio: pixelRatio);

      if (child.children != null) {
        children = [...children, ...child.children!];
      } else if (child.node != null) {
        children = [...children, child.node!];
      }
    }

    if (isElementPrivate(node.element)) {
      /// if element is private, we do not want to add a node for this corresponding element in the screengraph
      /// we directly return the children
      return _ScreengraphNodeOrChildren(children: children);
    }

    /// otherwise, we create a new node for the current element
    /// and return it's children recursively
    return _ScreengraphNodeOrChildren(node: {
      ...(await node.toScreengraphJson(pixelRatio: pixelRatio)),
      'children': children
    });
  }
}
