import 'package:contentsquare/src/modules/screengraph/screengraph_utils.dart';
import 'package:flutter/material.dart';
import 'package:contentsquare_platform_interface/contentsquare_platform_interface.dart';
import 'package:flutter/rendering.dart';
import 'package:flutter/widgets.dart';
import 'package:webview_flutter/webview_flutter.dart';

/// The format of the background color used here is: #OORRGGBB
/// OO: Opacity
/// RR: Red
/// GG: Green
/// BB: Blue
const _defaultElementBackgroundColor = '#00FFFFFF';

/// Represent a single node in the screengraph model
class ScreengraphNode {
  /// Parent [ScreengraphNode]
  final ScreengraphNode? parent;

  /// Flutter [Element] associated with the current [ScreengraphNode]
  final Element element;

  /// Screengraph path to the current node
  late String path;

  /// Constructor for the [ScreengraphNode] object
  ScreengraphNode({
    required this.element,
    this.parent,
    String? rootPath,
  }) : assert(parent != null || rootPath != null) {
    if (parent != null) {
      path = parent!.path;
    }
    if (rootPath != null) {
      path = rootPath;
    }

    if (!isElementPrivate(element)) {
      /// if the element is private, we avoid appending it to the path
      _appendPathWithElement(element);
    }
  }

  /// We append element with their positional index to be able to distinguish elements with the same path (sibling elements)
  /// By counting the occurences of each element path (starting at 0) we can attribute a positional index based on the number of occurences (minus 1)
  /// If an element is a single child (its element path will be unique) -> eq(0)
  /// If an element has 3 siblings (they will all have the same path), so we respectively attribute the positional indexes: eq(0), eq(1), eq(2)
  void appendPositionalIndexToPath(int positionalIndex) {
    if (!isElementPrivate(element)) {
      path = '$path:eq(${positionalIndex.toString()})';
    }
  }

  /// Visit all node's element's children and call the provided [onChildNodeVisited] callback for each
  void visitChildren(Function(ScreengraphNode node) onChildNodeVisited) {
    element.visitChildren((childElement) {
      ScreengraphNode childNode =
          ScreengraphNode(element: childElement, parent: this);
      onChildNodeVisited(childNode);
    });
  }

  /// Returns the frame Rect (position + size) of the element
  Rect get frame {
    final renderObject = element.findRenderObject();
    if (renderObject is RenderView) {
      final renderBox = renderObject.child;
      if (renderBox == null) {
        return Rect.zero;
      }
      return getFrameForBox(renderBox);
    } else if (renderObject is RenderBox) {
      return getFrameForBox(renderObject);
    } else if (renderObject is RenderSliver) {
      return getFrameForSliver(renderObject);
    } else if (renderObject is RenderAbstractViewport) {
      return getFrameForViewport(renderObject);
    } else {
      return Rect.zero;
    }
  }

  CSViewFormat get _viewFormat {
    /// For now we only support [CSViewFormat.mobileView] & [CSViewFormat.webViewContainer]
    /// TODO: Add support for CSViewFormat.webElement (https://contentsquare.atlassian.net/browse/FLUT-23)
    return (element.widget is WebView
        ? CSViewFormat.webViewContainer
        : CSViewFormat.mobileView);
  }

  Future<String?> get _bitmap async {
    return getElementBitmap(element);
  }

  /// Format current [ScreengraphNode] to a JSON map
  Future<Map> toScreengraphJson({required double pixelRatio}) async {
    final frame = this.frame;
    final bitmap = await _bitmap;
    return {
      'id': '${element.widget.key}',
      'metadata': {
        'class_name': element.widget.runtimeType.toString(),
        'fullpath': path,
      },

      /// "recordingId" should only mandatory for session replay
      /// On the iOS SDK, the structure is used for both: snapshoting and session replay features
      /// so it ends up mandatory for the screengraph passed down to the ExternalBridgeInterface.takeSnapshot completion
      /// For this reason, we set temporarily a 'recordingId' value based on the widget's hascode.
      /// TODO: This will need to be checked and discussed when implementing session replay related features
      'recordingId': element.widget.hashCode,
      'style': {
        'height': frame.height * pixelRatio,
        'width': frame.width * pixelRatio,
        'x': frame.left * pixelRatio,
        'y': frame.top * pixelRatio,
        'visibility': true,
        if (bitmap != null) 'bmp': bitmap,

        /// TODO: Retrieve the actual background color of the current element
        /// This requires to analyze the widget type and provide a way to retrieve the color for each type of widget
        if (bitmap == null) 'bg': _defaultElementBackgroundColor,
      },
      'format': _viewFormat.toInt(),
    };
  }

  void _appendPathWithElement(Element element) {
    path = '$path>${trimGenericParams(element.widget.runtimeType.toString())}';
  }
}
